@@ -1,9 +1,10 @@
 package PAR::Packer;
-use 5.006;
+use 5.008001;
 use strict;
 use warnings;
 
-our $VERSION = '1.000';
+our $VERSION = '1.029_01';
+$VERSION = eval $VERSION;
 
 =head1 NAME
 
@@ -27,7 +28,6 @@ have a C compiler.
 use Config;
 use Archive::Zip ();
 use Cwd ();
-use ExtUtils::MakeMaker ();
 use File::Basename ();
 use File::Find ();
 use File::Spec ();
@@ -47,11 +47,10 @@ use constant OPTIONS => {
     'e|eval:s'       => 'Packing one-liner',
     'E|evalfeature:s'=> 'Packing one-liner with new syntactic features',
     'x|execute'      => 'Execute code to get dependencies',
+    'xargs:s'        => 'Args to pass when executing code',
     'X|exclude:s@'   => 'Exclude modules',
     'f|filter:s@'    => 'Input filters for scripts',
     'g|gui'          => 'No console window',
-    'i|icon:s'       => 'Icon file',
-    'N|info:s@'      => 'Executable header info',
     'I|lib:s@'       => 'Include directories (for perl)',
     'l|link:s@'      => 'Include additional shared libraries',
     'L|log:s'        => 'Where to log packaging process information',
@@ -67,6 +66,7 @@ use constant OPTIONS => {
     'S|save'         => 'Preserve intermediate PAR files',
     's|sign'         => 'Sign PAR files',
     'T|tempcache:s'  => 'Temp cache name',
+    'u|unicode'      => 'Include Unicode stuff',
     'v|verbose:i'    => 'Verbosity level',
     'vv|verbose2',   => 'Verbosity level 2',
     'vvv|verbose3',  => 'Verbosity level 3',
@@ -191,9 +191,6 @@ sub _check_write {
         if (-e _) {
             $self->_die("Cannot write on $file: $!\n") unless -w _;
         }
-        unless (-w Cwd::cwd()) {
-            $self->_die("Cannot write in this directory: $!\n");
-        }
     }
 }
 
@@ -309,8 +306,8 @@ sub _parse_opts {
         $self->_warn("Using -e 'code' as input file, ignoring @$args\n")
           if (@$args and !$opt->{r});
 
-        my ($fh, $fake_input) =
-          File::Temp::tempfile("ppXXXXX", SUFFIX => ".pl", UNLINK => 1);
+        my ($fh, $fake_input) = File::Temp::tempfile(
+	    "ppXXXXX", SUFFIX => ".pl", TMPDIR => 1, UNLINK => 1);
 
         print $fh $opt->{e};
         close $fh;
@@ -320,7 +317,7 @@ sub _parse_opts {
         $self->{input} ||= [];
 
         # Reject main.pl as input file to avoid beginner confusion
-        if ( grep /main\.pl$/,
+        if ( grep /(?:^|[\/\\])main\.pl$/,
              ($opt->{r} ? ($args->[0]) : @$args) )
         {
           # -r means "run this" => extra args are execution parameters
@@ -361,7 +358,7 @@ sub _create_valid_hash {
     return () if (%$hashout);
 
     foreach my $key (keys(%$hashin)) {
-        my (@keys) = ($key =~ /\|/) ? ($key =~ /(?<!:)(\w+)/g) : ($key);
+        my (@keys) = $key =~ /(?<!:)(\w+)/g;
         @{$hashout}{@keys} = ($hashin->{$key}) x @keys;
     }
 }
@@ -548,6 +545,7 @@ sub _make_manifest {
     my ($self) = @_;
 
     my $full_manifest = $self->{full_manifest};
+    my $add_manifest  = $self->{add_manifest};
 
     my $opt      = $self->{options};
     my $par_file = $self->{par_file};
@@ -559,7 +557,7 @@ sub _make_manifest {
 
     my $manifest = join("\n",
 '    <!-- accessible as jar:file:///NAME.par!/MANIFEST in compliant browsers -->',
-        (sort keys %$full_manifest),
+        (sort keys %$full_manifest, keys %$add_manifest),
 q(    # <html><body onload="var X=document.body.innerHTML.split(/\n/);var Y='<iframe src=&quot;META.yml&quot; style=&quot;float:right;height:40%;width:40%&quot;></iframe><ul>';for(var x in X){if(!X[x].match(/^\s*#/)&&X[x].length)Y+='<li><a href=&quot;'+X[x]+'&quot;>'+X[x]+'</a>'}document.body.innerHTML=Y">)
     );
 
@@ -570,7 +568,7 @@ conflicts: {}
 dist_name: $dist_name
 distribution_type: par
 dynamic_config: 0
-generated_by: '$class version $version
+generated_by: '$class version $version'
 license: unknown
 par:
   clean: $clean
@@ -633,7 +631,8 @@ sub get_par_file {
         # Don't need to keep it, be safe with a tempfile.
 
         $self->{pack_attrib}{lose} = 1;
-        ($cfh, $par_file) = File::Temp::tempfile("ppXXXXX", SUFFIX => ".par");
+        ($cfh, $par_file) = File::Temp::tempfile(
+	    "ppXXXXX", SUFFIX => ".par", TMPDIR => 1, UNLINK => 1);
         close $cfh;    # See comment just below
     }
     $self->{par_file} = $par_file;
@@ -678,9 +677,18 @@ sub pack_manifest_hash {
     my (@modules, @data, @exclude);
 
     foreach my $name (@{ $opt->{M} || [] }) {
+        # for "-M Foo::Bar::" also add everything below .../Foo/Bar
+        if ($name =~ s/^([\w:]+)::$/$1/) {
+            (my $mod = $name) =~ s/::/\//g;
+            Module::ScanDeps::add_preload_rule("$mod.pm", 'sub');
+        }
         $self->_name2moddata($name, \@modules, \@data);
     }
 
+    if ($opt->{u}) {
+        push @modules, "utf8_heavy.pl";
+    }
+
     # Skip either
     # a) all files from a .par file or
     # b) A module
@@ -699,7 +707,7 @@ sub pack_manifest_hash {
     my %map;
 
     unshift(@INC, @{ $opt->{I} || [] });
-    unshift(@SharedLibs, map $self->_find_shlib($_, $sn), @{ $opt->{l} || [] });
+    unshift(@SharedLibs, map $self->_find_shlib($_), @{ $opt->{l} || [] });
 
     my $inc_find = $self->_obj_function($fe, '_find_in_inc');
 
@@ -721,22 +729,26 @@ sub pack_manifest_hash {
           or $self->_die("Cannot find module $module (specified with -M)\n");
         push @files, $file;
         
-        my $preload = Module::ScanDeps::_get_preload($module) or next;
+        my @preload = Module::ScanDeps::_get_preload($module) or next;
         
         $add_deps->(
             used_by => $file,
             rv      => \%map,
-            modules => $preload,
+            modules => \@preload,
             skip    => \%skip,
 #            warn_missing => $args->{warn_missing},
         );
-        push @files, map {&$inc_find($_)} @$preload;
     }
     push @files, @$input;
 
     # Search for scannable code in all -I'd paths
     push @Module::ScanDeps::IncludeLibs, @{$opt->{I}} if $opt->{I};
     
+    if ($opt->{x} && defined $opt->{xargs}) {
+        require Text::ParseWords;
+        $opt->{x} = [ Text::ParseWords::shellwords($opt->{xargs}) ];
+    }
+
     my $scan_dispatch =
       $opt->{n}
       ? $self->_obj_function($fe, 'scan_deps_runtime')
@@ -759,7 +771,7 @@ sub pack_manifest_hash {
     %skip = map { $_, 1 } map &$inc_find($_), @exclude;
     %skip = (%skip, map { $_, 1 } @SharedLibs);
 
-    &$add_deps(
+    $add_deps->(
         rv      => \%map,
         modules => \@modules,
         skip    => \%skip,
@@ -777,7 +789,7 @@ sub pack_manifest_hash {
     my $size = 0;
     my $old_member;
 
-    if ($opt->{'m'} and -e $par_file) {
+    if ($opt->{m} and -e $par_file) {
         my $tmpzip = Archive::Zip->new();
         $tmpzip->read($par_file);
 
@@ -814,21 +826,6 @@ sub pack_manifest_hash {
             $dep_manifest->{ $root . "lib/$pfile" } =
               [ string => $content_ref ];
         }
-        elsif (
-            File::Basename::basename($map{$pfile}) =~ /^Tk\.dll$/i and $opt->{i}
-            and eval { require Win32::Exe; 1 }
-            and eval { require Win32::Exe::IconFile; 1 }
-            and 0 # XXX - broken on larger icon files - XXX
-        ) {
-            my $tkdll = Win32::Exe->new($map{$pfile});
-            my $ico = Win32::Exe::IconFile->new($opt->{i});
-            $tkdll->set_icons(scalar $ico->icons);
-
-            $full_manifest->{ $root . "lib/$pfile" } =
-              [ string => $tkdll->dump ];
-            $dep_manifest->{ $root . "lib/$pfile" } =
-              [ string => $tkdll->dump ];
-        }
         else {
             $full_manifest->{ $root . "lib/$pfile" } =
               [ file => $map{$pfile} ];
@@ -855,6 +852,9 @@ sub pack_manifest_hash {
             $full_manifest->{"script/$name"} = [ file => $in ];
             $dep_manifest->{"script/$name"}  = [ file => $in ];
         }
+
+        delete $full_manifest->{$root . "lib/$name"};
+        delete $dep_manifest->{$root . "lib/$name"};
     }
 
     my $shlib = "shlib/$Config{archname}";
@@ -1110,9 +1110,9 @@ sub _add_file {
                 $manifest->{ $alias } = [ file => $file ];
 
                 $oldsize += -s $file;
-                $zip->addFile($file, $alias);
-                $zip->memberNamed($alias)->desiredCompressionMethod($method);
-                $zip->memberNamed($alias)->desiredCompressionLevel($level);
+                my $member = $zip->addFile($file, $alias);
+                $member->desiredCompressionMethod($method);
+                $member->desiredCompressionLevel($level);
             }
         }
         elsif (-e $fn and -r $fn) {
@@ -1123,9 +1123,9 @@ sub _add_file {
             $self->_vprint(1, "... adding $fn as $in\n");
 
             $oldsize += -s $fn;
-            $zip->addFile($fn => $in);
-            $zip->memberNamed($in)->desiredCompressionMethod($method);
-            $zip->memberNamed($in)->desiredCompressionLevel($level);
+            my $member = $zip->addFile($fn => $in);
+            $member->desiredCompressionMethod($method);
+            $member->desiredCompressionLevel($level);
         }
     }
     else {
@@ -1137,9 +1137,10 @@ sub _add_file {
         $oldsize += length($str);
 
         $self->_vprint(1, "... adding <string> as $in");
-        $zip->addString($str => $in)->unixFileAttributes(0644);
-        $zip->memberNamed($in)->desiredCompressionMethod($method);
-        $zip->memberNamed($in)->desiredCompressionLevel($level);
+        my $member = $zip->addString($str => $in);
+        $member->unixFileAttributes(0644);
+        $member->desiredCompressionMethod($method);
+        $member->desiredCompressionLevel($level);
     }
 
     $self->{pack_attrib}{old_size} = $oldsize;
@@ -1254,29 +1255,10 @@ sub _par_to_exe {
     if ($^O ne 'MSWin32' or $opt->{p} or $opt->{P}) {
         $self->_generate_output();
     }
-    elsif (!$opt->{N} and !$opt->{i}) {
+    else {
         $self->_generate_output();
         $self->_fix_console() if $opt->{g};
     }
-    elsif (eval { require Win32::Exe; 1 }) {
-        $self->_move_parl();
-        Win32::Exe->new($self->{parl})->update(
-            icon => $opt->{i},
-            info => $opt->{N},
-        );
-
-        $self->_append_parl();
-        $self->_generate_output();
-
-        $self->_fix_console();
-        unlink($self->{parl});
-        unlink($self->{orig_parl});
-        unlink("$self->{parl}.bak");
-        return;
-    }
-    else {
-        $self->_die("--icon and --info support needs Win32::Exe");
-    }
 }
 
 # extracts a parl (static) or parldyn (dynamic) from the appropriate data class
@@ -1296,9 +1278,8 @@ sub _extract_parl {
 
     $self->_vprint(0, "Generating a fresh 'parl'.");
     my ($fh, $filename) = File::Temp::tempfile(
-        "parlXXXXXXX",
-        SUFFIX => $Config{_exe},
-    );
+        "parlXXXXXXX", SUFFIX => $Config{_exe}, TMPDIR => 1, UNLINK => 1);
+    close $fh;
     
     my $success = $class->write_parl($filename);
     if (not $success) {
@@ -1310,34 +1291,6 @@ sub _extract_parl {
 }
 
 
-sub _fix_console {
-    my ($self) = @_;
-
-    my $opt      = $self->{options};
-    my $output   = $self->{output};
-    my $dynperl  = $self->{dynperl};
-
-    return unless $opt->{g};
-
-    $self->_vprint(1, "Fixing $output to remove its console window");
-    $self->_strip_console($output);
-
-    if ($dynperl and !$opt->{d}) {
-        # we have a static.exe that needs taking care of.
-        my $buf;
-        my $fh = $self->_open($self->{orig_parl} || $self->{parl});
-        seek $fh, -8, 2;
-        read $fh, $buf, 8;
-        die unless $buf eq "\nPAR.pm\n";
-        seek $fh, -12, 2;
-        read $fh, $buf, 4;
-        seek $fh, -12 - unpack("N", $buf) - 4, 2;
-        read $fh, $buf, 4;
-        close $fh;
-        $self->_strip_console($output, unpack("N", $buf));
-    }
-}
-
 sub _move_parl {
     my ($self) = @_;
 
@@ -1346,10 +1299,7 @@ sub _move_parl {
     my $cfh;
     my $fh = $self->_open($self->{parl});
     ($cfh, $self->{parl}) = File::Temp::tempfile(
-        "parlXXXX",
-        SUFFIX => ".exe",
-        UNLINK => 1,
-    );
+        "parlXXXX", SUFFIX => ".exe", TMPDIR => 1, UNLINK => 1);
     binmode($cfh);
 
     local $/;
@@ -1359,25 +1309,6 @@ sub _move_parl {
     $self->{fh} = $fh;
 }
 
-sub _append_parl {
-    my ($self) = @_;
-
-    my $fh = delete $self->{fh};
-
-    my $buf;
-    seek $fh, -8, 2;
-    read $fh, $buf, 8;
-    die unless $buf eq "\nPAR.pm\n";
-    seek $fh, -12, 2;
-    read $fh, $buf, 4;
-    seek $fh, -12 - unpack("N", $buf), 2;
-
-    local $/;
-    my $cfh = $self->_open('>>', $self->{parl});
-    print $cfh readline($fh);
-    close $cfh;
-}
-
 sub _generate_output {
     my ($self) = @_;
 
@@ -1412,26 +1343,32 @@ sub _generate_output {
     system($parl, @args);
 }
 
-sub _strip_console {
-    my $self   = shift;
-    my $file   = shift;
-    my $preoff = shift || 0;
+sub _fix_console {
+    my ($self) = @_;
+
+    my $opt      = $self->{options};
+    my $output   = $self->{output};
+    my $dynperl  = $self->{dynperl};
+
+    return unless $opt->{g};
+
+    $self->_vprint(1, "Fixing $output to remove its console window");
 
     my ($record, $magic, $signature, $offset, $size);
 
-    my $exe = $self->_open('+<', $file);
+    my $exe = $self->_open('+<', $output);
     binmode $exe;
-    seek $exe, $preoff, 0;
+    seek $exe, 0, 0;
 
     # read IMAGE_DOS_HEADER structure
     read $exe, $record, 64;
     ($magic, $offset) = unpack "Sx58L", $record;
 
-    die "$file is not an MSDOS executable file.\n"
+    die "$output is not an MSDOS executable file.\n"
       unless $magic == 0x5a4d;    # "MZ"
 
     # read signature, IMAGE_FILE_HEADER and first WORD of IMAGE_OPTIONAL_HEADER
-    seek $exe, $preoff + $offset, 0;
+    seek $exe, $offset, 0;
     read $exe, $record, 4 + 20 + 2;
 
     ($signature, $size, $magic) = unpack "Lx16Sx2S", $record;
@@ -1444,7 +1381,7 @@ sub _strip_console {
       ($size == 240 && $magic == 0x20b);    # IMAGE_NT_OPTIONAL_HDR64_MAGIC
 
     # Offset 68 in the IMAGE_OPTIONAL_HEADER(32|64) is the 16 bit subsystem code
-    seek $exe, $preoff + $offset + 4 + 20 + 68, 0;
+    seek $exe, $offset + 4 + 20 + 68, 0;
     print $exe pack "S", 2;                 # IMAGE_WINDOWS
     close $exe;
 }
@@ -1506,6 +1443,8 @@ sub _check_par {
 sub _chase_lib {
    my ($self, $file) = @_;
 
+   return $self->_chase_lib_darwin($file) if $^O eq q/darwin/;
+
    while ($Config::Config{d_symlink} and -l $file) {
        if ($file =~ /^(.*?\.\Q$Config{dlext}\E\.\d+)\..*/) {
            return $1 if -e $1;
@@ -1528,8 +1467,36 @@ sub _chase_lib {
    return $file;
 }
 
+sub _chase_lib_darwin {
+   my ($self, $file) = @_;
+
+   while (-l $file) {
+       if ($file =~ /^(.*?\.\d+)(\.\d+)*\.dylib$/) {
+           my $name = $1 . q/.dylib/;
+           return $name if -e $name;
+       }
+
+       return $file if $file =~ /\D\.\d+\.dylib$/;
+
+       my $dir = File::Basename::dirname($file);
+       $file = readlink($file);
+
+       unless (File::Spec->file_name_is_absolute($file)) {
+           $file = File::Spec->rel2abs($file, $dir);
+       }
+   }
+
+   if ($file =~ /^(.*?\.\d+)(\.\d+)*\.dylib$/) {
+       my $name = $1 . q/.dylib/;
+       return $name if -e $name;
+   }
+
+   return $file;
+}
+
+
 sub _find_shlib {
-    my ($self, $file, $script_name) = @_;
+    my ($self, $file) = @_;
 
     if ($^O eq 'MSWin32') {
         if ($file !~ /^[a-z]:/i) {
@@ -1550,25 +1517,41 @@ sub _find_shlib {
                            ? $ENV{ $Config{ldlibpthname} } : undef;
     }
     if (not defined $libpthname) {
-        print "Can't find $file. Environment variable "
-          . ($^O eq 'MSWin32' ? 'PATH' : $Config{ldlibpthname})
-          . " does not exist.\n";
+        $self->_die(sprintf(
+                "Don't know how to find shared library (option -l) %s: ".
+                "environment variable %s is not set.",
+                $file, $^O eq 'MSWin32' ? 'PATH' : $Config{ldlibpthname}));
         return;
     }
 
     $file = File::Basename::basename($file);
-    for my $dir (File::Basename::dirname($0),
-        split(/\Q$Config{path_sep}\E/, $libpthname))
-    {
-        my $abs = File::Spec->catfile($dir, $file);
-        return $self->_chase_lib($abs) if -e $abs;
-        $abs = File::Spec->catfile($dir, "$file.$Config{dlext}");
-        return $self->_chase_lib($abs) if -e $abs;
-    }
+    my @path = (File::Basename::dirname($0), split(/\Q$Config{path_sep}\E/, $libpthname));
+
+    my $lib = $self->_find_shlib_in_path($file, @path);
+    return $lib if $lib;
+
+    $self->_die("Can't find shared library (option -l): $file")
+        if $^O eq 'MSWin32' || $file =~ /^lib/;
 
     # be extra magical and prepend "lib" to the filename
-    return if $^O eq 'MSWin32';
-    return $self->_find_shlib("lib$file", $script_name) unless $file =~ /^lib/;
+    $lib = $self->_find_shlib_in_path("lib$file", @path);
+    return $lib if $lib;
+
+    $self->_die("Can't find shared library (option -l): $file (also tried lib$file)");
+}
+
+sub _find_shlib_in_path
+{
+    my ($self, $file, @path) = @_;
+
+    my $dlext = $^O eq 'darwin' ? 'dylib' : $Config{dlext};
+    for my $dir (@path)
+    {
+        my $abs = File::Spec->catfile($dir, $file);
+        $abs = File::Spec->catfile($dir, "$file.$dlext") unless -e $abs;
+        return $self->_chase_lib($abs) if -e $abs;
+    }
+    return;
 }
 
 sub _can_run {
@@ -1681,17 +1664,22 @@ sub _main_pl_clean {
     my $clean_inc = '';
     if ($opt->{B}) { # bundle core modules
         # weed out all @INC entries
+        # use a canonicalized $ENV{PAR_TEMP}: this path was created by C code
+        # and may not be in canonical form (so that the match below will
+        # fail); case inpoint: some versions of FreeBSD have
+        #  #define P_tmpdir "/var/tmp/"
+        # in /usr/include/stdio.h (note the trailing slash)
         $clean_inc = <<'__CLEAN_INC__';
 # Remove everything but PAR hooks from @INC
 my %keep = (
     \&PAR::find_par => 1,
     \&PAR::find_par_last => 1,
 );
-my $par_temp_dir = quotemeta( $ENV{PAR_TEMP} );
+my $par_temp_dir = File::Spec->catdir( $ENV{PAR_TEMP} );
 @INC =
     grep {
         exists($keep{$_})
-        or $_ =~ /^$par_temp_dir/;
+        or $_ =~ /^\Q$par_temp_dir\E/;
     }
     @INC;
 __CLEAN_INC__
@@ -1732,7 +1720,7 @@ Audrey Tang E<lt>cpan@audreyt.orgE<gt>
 
 Steffen Mueller E<lt>smueller@cpan.orgE<gt>
 
-L<http://par.perl.org/> is the official PAR website.  You can write
+You can write
 to the mailing list at E<lt>par@perl.orgE<gt>, or send an empty mail to
 E<lt>par-subscribe@perl.orgE<gt> to participate in the discussion.
 
@@ -1740,7 +1728,7 @@ Please submit bug reports to E<lt>bug-par@rt.cpan.orgE<gt>.
 
 =head1 COPYRIGHT
 
-Copyright 2004-2009 by Audrey Tang E<lt>cpan@audreyt.orgE<gt>.
+Copyright 2004-2010 by Audrey Tang E<lt>cpan@audreyt.orgE<gt>.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
